@@ -28,6 +28,16 @@ class PluginEntryPoint(object):
     PREFIX_FREE_DISTRIBUTIONS = [
         "certbot",
         "certbot-apache",
+        "certbot-dns-cloudflare",
+        "certbot-dns-cloudxns",
+        "certbot-dns-digitalocean",
+        "certbot-dns-dnsimple",
+        "certbot-dns-dnsmadeeasy",
+        "certbot-dns-google",
+        "certbot-dns-luadns",
+        "certbot-dns-nsone",
+        "certbot-dns-rfc2136",
+        "certbot-dns-route53",
         "certbot-nginx",
     ]
     """Distributions for which prefix will be omitted."""
